unit dijkAlg;

interface

uses tGraf, tPrVrsta;

const maxRazdalja=999; //razdalja vozlisca od zacetnega na zacetku izvajanja Dijkstrovega algoritma (namesto neskoncno)

type
  dijkVoz=class;
  dijkEl=class (element)
    voz: dijkVoz; //vozlisce grafa
  end;
  dijkPrVrsta=class (prVrsta)
    procedure dodaj (el: dijkEl); //doda element <el>
  end;
  dijkVoz=class (vozlisce)
    el: dijkEl; //element prioritetne vrste, v katerem je to vozlisce
    odZac: real; //razdalja od zacetnega elementa
    predhodnik: dijkVoz; //vozlisce, od katerega vodi najkrajsa povezava do tega
    constructor create;
  end;
  dijkGraf=class (graf)
    vrsta: dijkPrVrsta; //prioritetna vrsta, ki jo uporablja dijkstraKorak
    zadnjeVoz: vozlisce; //zadnje vozlisce, s katerim se je ukvarjal dijkstraKorak
    procedure trenutno (trenutnoVoz: dijkVoz; trenutnaPov: povezava); virtual; abstract; //stori nekaj, ko se program ukvarja z vozliscem <trenutnoVoz> in povezavo <trenutnaPov>
    procedure dijkstraCel (zac: vozlisce); //zacensi v vozliscu <zac> izvede Dijkstrov algoritem
    function dijkstraKorak (zac: vozlisce): boolean; //zacensi v vozliscu <zac> izvede Dijkstrov algoritem po korakih; dokler je potreben se kak korak, vraca true, potem pa false
  end;

implementation

constructor dijkVoz.create;
begin
  inherited;
  odZac:=maxRazdalja; //razdaljo od zacetnega vozlisca nastavi na 'neskoncno'
end;

procedure dijkPrVrsta.dodaj (el: dijkEl);
begin
  el.prioriteta:=el.voz.odZac; //prioriteto priredi razdaljo pripadajocega vozlisca od zacetnega
  el.voz.el:=el; //pripadajocemu element doda kazalec nase (zaradi znizevanje prioritete)
  inherited dodaj (el);
end;

procedure dijkGraf.dijkstraCel (zac: vozlisce);
  var
    vrsta: dijkPrVrsta;
    el: dijkEl;
    voz: vozlisce;
    pov: povezava;
begin
  vrsta:=dijkPrVrsta.create;
  voz:=prvoVoz;
  while voz<>nil do begin //gre po vseh vozliscih
    if voz<>zac then dijkVoz (voz).odZac:=maxRazdalja else dijkVoz (voz).odZac:=0;
    el:=dijkEl.create;
    el.voz:=dijkVoz (voz);
    vrsta.dodaj (el); //element s trenutnim vozliscem da v prioritetno vrsto
    voz:=dijkVoz (voz.naslednje);
  end;
  while vrsta.stEl>0 do begin //dokler je v vrsti se kaj
    el:=dijkEl (vrsta.brisiMin); //zbrise tistega z najmanjso razdaljo od zacetnega vozlisce
    pov:=el.voz.prvaPov;
    while pov<>nil do begin //gre po vseh njegovih povezavah
      if dijkVoz (pov.kam).odZac>el.voz.odZac+pov.dolzina then begin //ce je pod do vozlisce na koncu trenutne povezave prek trenutnega vozlisce manjsa od najmanjse do sedaj znane
        dijkVoz (pov.kam).odZac:=el.voz.odZac+pov.dolzina; //popravi razdaljo
        dijkVoz (pov.kam).predhodnik:=el.voz; //popravi predhodnika (ker je pot prek trenutnega vozlisca krajsa)
        vrsta.znizajPr (dijkVoz (pov.kam).el, dijkVoz (pov.kam).odZac); //zaradi spremenjene razdalje spremeni prioriteto
        trenutno (el.voz, pov);
      end;
      pov:=pov.naslednja;
    end;
  end;
  vrsta.free;
end;

function dijkGraf.dijkstraKorak (zac: vozlisce): boolean; //glej prejsnjo metodo
  var //namesto lokalne vrste uporablja grafovo
    voz: vozlisce;
    el: dijkEl;
    pov: povezava;
begin
  if vrsta=nil then begin
    vrsta:=dijkPrVrsta.create;
    voz:=prvoVoz;
    while voz<>nil do begin
      if voz<>zac then dijkVoz (voz).odZac:=maxRazdalja else dijkVoz (voz).odZac:=0;
      el:=dijkEl.create;
      el.voz:=dijkVoz (voz);
      vrsta.dodaj (el);
      voz:=dijkVoz (voz.naslednje);
    end;
  end;
  if vrsta.stEl>0 then begin //namesto da bi iz vrste brisal v zanki, brise po enega na enkrat
    el:=dijkEl (vrsta.brisiMin);
    pov:=el.voz.prvaPov;
    while pov<>nil do begin
      if dijkVoz (pov.kam).odZac>el.voz.odZac+pov.dolzina then begin
        dijkVoz (pov.kam).odZac:=el.voz.odZac+pov.dolzina;
        dijkVoz (pov.kam).predhodnik:=el.voz;
        vrsta.znizajPr (dijkVoz (pov.kam).el, dijkVoz (pov.kam).odZac);
      end;
      pov:=pov.naslednja;
    end;
    zadnjeVoz:=el.voz;
  end;
  if vrsta.stEl=0 then begin //doloci, ce je potreben se kak korak
    vrsta.free;
    vrsta:=nil;
    dijkstraKorak:=false;
  end else dijkstraKorak:=true;
end;

end.
